package com.tngtech.jgiven.impl.tag;


import static org.assertj.core.api.Assertions.assertThat;

import com.tngtech.jgiven.config.DefaultConfiguration;
import com.tngtech.jgiven.impl.TestUtil.JGivenLogHandler;
import com.tngtech.jgiven.report.model.Tag;
import java.lang.annotation.Annotation;
import java.util.Arrays;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.LogManager;
import java.util.logging.Logger;
import org.junit.Before;
import org.junit.Test;

public class TagCreatorTest {

    private final TagCreator underTest = new TagCreator(new DefaultConfiguration());
    private final JGivenLogHandler interceptor = new JGivenLogHandler();

    @Before
    public void addLogInterceptor() {
        Logger testLogger = LogManager.getLogManager().getLogger(TagCreator.class.getName());
        testLogger.addHandler(interceptor);
    }

    @Test
    public void testAnnotationParsing() {
        Tag tag = getOnlyTagFor(AnnotationTestClass.class.getAnnotations()[0]);
        assertThat(tag.getName()).isEqualTo("AnnotationWithoutValue");
        assertThat(tag.getValues()).isEmpty();
        assertThat(interceptor.containsLoggingEvent(record -> record.getLevel() == Level.SEVERE))
            .as("Attempt to convert an annotation without value method results in an error log")
            .isFalse();
    }

    @Test
    public void testAnnotationWithValueParsing() {
        Tag tag = getOnlyTagFor(AnnotationWithSingleValueTestClass.class.getAnnotations()[0]);
        assertThat(tag.getName()).isEqualTo("AnnotationWithSingleValue");
        assertThat(tag.getValues()).containsExactly("testvalue");
    }

    @Test
    public void testAnnotationWithName() {
        Tag tag = getOnlyTagFor(AnnotationWithNameTestClass.class.getAnnotations()[0]);
        assertThat(tag.getName()).isEqualTo("AnotherName");
        assertThat(tag.getValues()).isEmpty();
        assertThat(tag.toIdString()).isEqualTo(AnnotationWithName.class.getName());
    }

    @Test
    public void testAnnotationWithIgnoredValueParsing() {
        Tag tag = getOnlyTagFor(AnnotationWithIgnoredValueTestClass.class.getAnnotations()[0]);
        assertThat(tag.getName()).isEqualTo("AnnotationWithIgnoredValue");
        assertThat(tag.getValues()).isEmpty();
        assertThat(tag.toIdString()).isEqualTo(AnnotationWithIgnoredValue.class.getName());
    }

    @Test
    public void testAnnotationWithoutExplodedArrayParsing() {
        Tag tag = getOnlyTagFor(AnnotationWithoutExplodedArrayValueTestClass.class.getAnnotations()[0]);
        assertThat(tag.getName()).isEqualTo("AnnotationWithoutExplodedArray");
        assertThat(tag.getValues()).containsExactly("foo", "bar");
    }

    @Test
    public void testAnnotationWithDescription() {
        Tag tag = getOnlyTagFor(AnnotationWithDescription.class.getAnnotations()[0]);
        assertThat(tag.getDescription()).isEqualTo("Some Description");
    }

    @Test
    public void testAnnotationWithDescriptionAndIgnoreValue() {
        Tag tag = getOnlyTagFor(AnnotationWithDescriptionAndIgnoreValue.class.getAnnotations()[0]);
        assertThat(tag.getValues()).isEmpty();
        assertThat(tag.getDescription()).isEqualTo("Some Description");
        assertThat(tag.getTags()).hasSize(1);
    }

    @Test
    public void testAnnotationWithParentTag() {
        Tag tag = getOnlyTagFor(AnnotationWithParentTag.class.getAnnotations()[0]);
        assertThat(tag.getTags()).containsAll(Arrays.asList(
            ParentTag.class.getName(),
            ParentTagWithValue.class.getPackage().getName() + ".ParentTagWithValue-SomeValue")
        );
    }

    @Test
    public void testAnnotationWithArrayParsing() {
        List<Tag> tags =
            underTest.toTags(AnnotationWithArrayValueTestClass.class.getAnnotations()[0]).getDeclaredTags();
        assertThat(tags).hasSize(2);
        assertThat(tags.get(0).getName()).isEqualTo("AnnotationWithArray");
        assertThat(tags.get(0).getValues()).containsExactly("foo");
        assertThat(tags.get(1).getName()).isEqualTo("AnnotationWithArray");
        assertThat(tags.get(1).getValues()).containsExactly("bar");
    }

    @Test
    public void testAllParentsOfTagAreResolved() {
        ResolvedTags resolvedTags = underTest.toTags(TagWithGrandparentTags.class);
        assertThat(resolvedTags.getDeclaredTags())
            .extracting(Tag::getName)
            .containsExactly("TagWithGrandparentTags");
        assertThat(resolvedTags.getParents())
            .extracting(Tag::getName)
            .containsExactlyInAnyOrder("TagWithParentTags", "ParentTag", "ParentTagWithValue");
    }

    private Tag getOnlyTagFor(Annotation annotation) {
        List<Tag> tags = underTest.toTags(annotation).getDeclaredTags();
        assertThat(tags).hasSize(1);
        return tags.get(0);
    }
}
